package tools.descartes.prisma.agent.generic;

import java.io.File;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.equinox.app.IApplication;
import org.eclipse.equinox.app.IApplicationContext;

import edu.kit.ipd.descartes.mm.resourcetype.ResourceType;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceTypeRepository;
import edu.kit.ipd.descartes.mm.resourcetype.ResourcetypePackage;
import tools.descartes.prisma.agent.generic.genericagent.DelegationBinding;
import tools.descartes.prisma.agent.generic.genericagent.GenericAgentConfiguration;
import tools.descartes.prisma.agent.generic.genericagent.RequiredObject;
import tools.descartes.prisma.core.MessageBus;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ObjectMatcher;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.core.impl.RabbitMQMessageBus;
import tools.descartes.prisma.core.scopes.ModelNotification;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class GenericAgent implements IApplication {

	private class GenericAgentController extends AgentController {
		private final List<DelegationBinding> delegations;
		private final Set<EObject> delegatedObjects = new HashSet<>();

		public GenericAgentController(ModelRepository repository, MessageBus messageBus,
				GenericAgentConfiguration configuration) {
			super(repository, messageBus, configuration.getScope().getPath(), configuration.getId());
			delegations = configuration.getDelegations();
		}

		@Override
		public void handleNotification(ModelNotification notification) {
			if (notification.getNotificationType().equals(ModelNotification.NEW_NOTIFICATION)) {
				delegatedObjects.add(notification.getInstance());
			} else if (notification.getNotificationType().equals(ModelNotification.REMOVED_NOTIFICATION)) {
				delegatedObjects.remove(notification.getInstance());
			}
		}

		public Set<EObject> getDelegatedObjects() {
			return delegatedObjects;
		}

		@Override
		protected void startAsync() {
			super.startAsync();

			for (DelegationBinding binding : delegations) {
				String prefix = binding.getRoutingPrefix();
				if (prefix == null) {
					prefix = "";
				}
				getContainerScope().addDelegationBinding(binding.getSourcePort(),
						binding.getDestinationScope().getPath(), binding.getTargetPort(),
						prefix);
			}
		}

	}

	@Override
	public Object start(IApplicationContext context) throws Exception {
		String[] args = (String[]) context.getArguments().get(IApplicationContext.APPLICATION_ARGS);

		Options options = new Options();
		options.addOption("r", "repository", true, "Host of the model repository");
		options.addOption("u", "user", true, "User of the model repository");
		options.addOption("p", "password", true, "Password of the model repository");
		options.addOption("c", "config", true, "Agent configuration");

		CommandLineParser cliParser = new GnuParser();
		CommandLine cli = cliParser.parse(options, args);

		String repositoryHost = null;
		String repositoryUser = null;
		String repositoryPassword = null;
		String configurationPath = null;
		boolean help = false;
		if (cli.hasOption("r")) {
			repositoryHost = cli.getOptionValue("r");
		} else {
			help = true;
		}
		if (cli.hasOption("u")) {
			repositoryUser = cli.getOptionValue("u");
		} else {
			help = true;
		}
		if (cli.hasOption("p")) {
			repositoryPassword = cli.getOptionValue("p");
		} else {
			help = true;
		}
		if (cli.hasOption("c")) {
			configurationPath = cli.getOptionValue("c");
		} else {
			help = true;
		}

		if (help) {
			HelpFormatter helpFormat = new HelpFormatter();
			helpFormat.printHelp("genericagent", options);
			return null;
		}
		GenericAgentConfiguration configuration = loadConfiguration(configurationPath);

		MessageBus messageBus = new RabbitMQMessageBus(repositoryHost, repositoryUser, repositoryPassword);
		ModelRepository repository = new CDOModelRepository(repositoryHost, repositoryUser, repositoryPassword,
				messageBus);
		GenericAgentController agent = new GenericAgentController(repository, messageBus, configuration);
		agent.start().get();

		resolvePublicResources(agent, configuration.getSkeleton());

		if (resolveRequiredObjects(agent, configuration)) {
			agent.apply(configuration.getSkeleton());
		}
		
		Thread.sleep(20000);

		agent.stop().get();

		return null;
	}

	@Override
	public void stop() {
		// TODO Auto-generated method stub
	}

	private GenericAgentConfiguration loadConfiguration(String configurationPath) {
		ResourceSet set = new ResourceSetImpl();
		Resource res = set.getResource(URI.createFileURI(new File(configurationPath).getAbsolutePath()), true);
		GenericAgentConfiguration configuration = (GenericAgentConfiguration) res.getContents().get(0);

		// IMPORTANT: resolve all cross-references
		EcoreUtil.resolveAll(set);
		Map<EObject, Collection<Setting>> unresolved = EcoreUtil.UnresolvedProxyCrossReferencer.find(set);
		for (EObject problemObject : unresolved.keySet()) {
			Collection<Setting> unresolvedSettings = unresolved.get(problemObject);
			for (Setting curSetting : unresolvedSettings) {
				System.out.println("Unresolved proxy: " + curSetting.getEStructuralFeature());
			}
		}

		return configuration;
	}

	private void resolvePublicResources(GenericAgentController agent, ModelSkeleton skeleton) {
		// Resource types (e.g., CPU, HDD, DELAY) which are defined globally are
		// not included in the model skeleton.
		// We need to find these references in the model skeleton and connect
		// them with the instances in the repository.
		TreeIterator<EObject> iterator = skeleton.eAllContents();
		ResourceTypeRepository publicResourceTypes = agent.getContainerScope().getSystemScope().getResourceTypes();
		while (iterator.hasNext()) {
			EObject curObject = iterator.next();
			for (EReference curReference : curObject.eClass().getEAllReferences()) {
				if (!curReference.isContainment() && !curReference.isContainer()) {
					if (ResourcetypePackage.Literals.RESOURCE_TYPE.isSuperTypeOf(curReference.getEReferenceType())) {
						ResourceType curType = (ResourceType) curObject.eGet(curReference);
						if (curType != null) {
							// find matching resources in system global scope
							// (matching is based on name and class
							for (ResourceType curPublicOne : publicResourceTypes.getResourceTypes()) {
								if (curType.eClass().isInstance(curPublicOne)
										&& curType.getName().equals(curPublicOne.getName())) {
									ModelSkeletonAdapter adapter = agent.adapt(curType, ModelSkeletonAdapter.class);
									adapter.setRepositoryObject(curPublicOne);
									break;
								}
							}
						}
					}
				}
			}
		}
	}


	private boolean resolveRequiredObjects(GenericAgentController agent, GenericAgentConfiguration configuration) {
		Set<RequiredObject> notResolved = new HashSet<>(configuration.getRequiredObjects());
		for (EObject curDelegate : agent.getDelegatedObjects()) {
			for (RequiredObject reqObject : notResolved) {
				if (ObjectMatcher.matches(curDelegate, reqObject.getObject())) {
					// Connect the two objects with each other
					ModelSkeletonAdapter adapter = agent.adapt(reqObject.getObject(),
							ModelSkeletonAdapter.class);
					adapter.setRepositoryObject((CDOObject) curDelegate);

					// Requirement is fulfilled
					notResolved.remove(reqObject);
					break;
				}
			}
		}

		if (!notResolved.isEmpty()) {
			for (RequiredObject unresolvedObject : notResolved) {
				System.err.println("Could not resolve required object: " + unresolvedObject.getObject());
			}
			return false;
		}
		return true;
	}

}
