package tools.descartes.prisma.agent.wildfly.callstack;

import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import tools.descartes.prisma.agent.wildfly.service.ModelExtractionService;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

public class InvocationRecord {

	// Invocation statistics
	public final AtomicLong invocationCount = new AtomicLong();
	public final AtomicLong executionTime = new AtomicLong();
	public final AtomicLong cpuTime = new AtomicLong();
	public final AtomicLong beforeCpuTime = new AtomicLong();
	public final ConcurrentHashMap<String, OperationRecord> calledEndpoints = new ConcurrentHashMap<String, OperationRecord>();

	private final MonitoringService monitoringService;
	private final OperationRecord operation;
	private final int valueBucket;
	private String stringRepresentation;

	public InvocationRecord(MonitoringService monitoringService, OperationRecord operation, int valueBucket) {
		this.monitoringService = monitoringService;
		this.operation = operation;
		this.valueBucket = valueBucket;
	}

	public int getValueBucket() {
		return valueBucket;
	}

	public OperationRecord getOperation() {
		return operation;
	}

	public long getInvocationCount() {
		return invocationCount.get();
	}

	public long getExecutionTime() {
		return executionTime.get();
	}

	public String getAdditionalInformation() {
		return String.format("%d, %d, %d", executionTime.get() - getExecutionTimeOfCalls(), cpuTime.get(),
				beforeCpuTime.get());
	}

	public void finishInvocation(long timeInNanos, long cpuTimeInNanos, long beforeCpuTimeInNanos) {
		if (invocationCount.incrementAndGet() == 1) {
			ModelExtractionService.INSTANCE.newInvocationRecord(this);
		}
		executionTime.addAndGet(timeInNanos);
		cpuTime.addAndGet(cpuTimeInNanos);
		beforeCpuTime.addAndGet(beforeCpuTimeInNanos);
	}

	private long getExecutionTimeOfCalls() {
		long total = 0;
		for (OperationRecord reqOrec : calledEndpoints.values()) {
			double sum = 0.0;
			long totalWheight = 0;
			for (InvocationRecord invocation : reqOrec.getInvocations()) {
				long count = invocation.getInvocationCount();
				sum += count * invocation.getExecutionTime();
				totalWheight += count;
			}
			if (totalWheight > 0) {
				total += (sum / totalWheight);
			}
		}
		return total;
	}

	public OperationRecord requireEndpoint(ComponentRecord component, String interf, String protocol,
			String operation) {
		String endpointIdentifier = OperationRecord.getEndpointIdentification(component, interf, protocol, operation);
		OperationRecord record = calledEndpoints.get(endpointIdentifier);
		if (record == null) {
			record = monitoringService.createOperationRecord(this, component, interf, protocol, operation);
			if (calledEndpoints.putIfAbsent(endpointIdentifier, record) != null) {
				// a different thread has changed it in the mean time
				record = calledEndpoints.get(endpointIdentifier);
			} else {
				ModelExtractionService.INSTANCE.newOperationRecord(record);
			}
		}
		return record;
	}

	public Collection<OperationRecord> getCalledEndpoints() {
		return calledEndpoints.values();
	}

	@Override
	public String toString() {
		if (stringRepresentation == null) {
			StringBuilder res = new StringBuilder();
			res.append(operation);
			if (operation.areParametersMonitored()) {
				res.append("?bucket=");
				res.append(valueBucket);
			}
			stringRepresentation = res.toString();
		}
		return stringRepresentation;
	}

}
