package tools.descartes.prisma.agent.wildfly.deployment;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jboss.as.ee.component.Attachments;
import org.jboss.as.ee.component.ComponentConfiguration;
import org.jboss.as.ee.component.ComponentConfigurator;
import org.jboss.as.ee.component.ComponentDescription;
import org.jboss.as.ee.component.EEModuleDescription;
import org.jboss.as.ee.component.ViewConfiguration;
import org.jboss.as.ejb3.component.EJBComponentDescription;
import org.jboss.as.ejb3.component.EJBViewConfiguration;
import org.jboss.as.server.AbstractDeploymentChainStep;
import org.jboss.as.server.deployment.DeploymentPhaseContext;
import org.jboss.as.server.deployment.DeploymentUnit;
import org.jboss.as.server.deployment.DeploymentUnitProcessingException;
import org.jboss.as.server.deployment.DeploymentUnitProcessor;
import org.jboss.as.server.deployment.Phase;
import org.jboss.as.web.common.WarMetaData;
import org.jboss.as.webservices.injection.WSComponentDescription;
import org.jboss.as.webservices.metadata.model.AbstractEndpoint;
import org.jboss.as.webservices.metadata.model.EJBEndpoint;
import org.jboss.as.webservices.metadata.model.JAXWSDeployment;
import org.jboss.as.webservices.metadata.model.POJOEndpoint;
import org.jboss.as.webservices.util.WSAttachmentKeys;
import org.jboss.invocation.ImmediateInterceptorFactory;
import org.jboss.logging.Logger;
import org.jboss.metadata.web.spec.ServletMappingMetaData;
import org.jboss.metadata.web.spec.ServletMetaData;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.interceptors.EJBLocalCallInterceptor;
import tools.descartes.prisma.agent.wildfly.interceptors.EJBMessageInterceptor;
import tools.descartes.prisma.agent.wildfly.interceptors.EJBRemoteCallInterceptor;
import tools.descartes.prisma.agent.wildfly.interceptors.EJBWebServiceCallInterceptor;
import tools.descartes.prisma.agent.wildfly.interceptors.WebServiceCallInterceptor;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

/**
 * An example deployment unit processor that does nothing. To add more deployment
 * processors copy this class, and add to the {@link AbstractDeploymentChainStep}
 * {@link tools.descartes.prisma.agent.wildfly.extension.SubsystemAdd#performBoottime(org.jboss.as.controller.OperationContext, org.jboss.dmr.ModelNode, org.jboss.dmr.ModelNode, org.jboss.as.controller.ServiceVerificationHandler, java.util.List)}
 *
 * @author <a href="simon.spinner@uni-wuerzburg.de">Simon Spinner</a>
 */
public class ServerInstrumentationDeploymentProcessor implements DeploymentUnitProcessor {

    Logger log = Logger.getLogger(ServerInstrumentationDeploymentProcessor.class);

    /**
     * See {@link Phase} for a description of the different phases
     */
    public static final Phase PHASE = Phase.POST_MODULE;

    /**
     * The relative order of this processor within the {@link #PHASE}.
     * The current number is large enough for it to happen after all
     * the standard deployment unit processors that come with JBoss AS.
     */
    public static final int PRIORITY = 0x4000;

    @Override
    public void deploy(DeploymentPhaseContext phaseContext) throws DeploymentUnitProcessingException {
    	DeploymentUnit deploymentUnit = phaseContext.getDeploymentUnit();
    	final MonitoringService monitoringService = MonitoringService.INSTANCE;
        log.info("Instrumenting deployment unit: " + deploymentUnit.getName());

		final EEModuleDescription eeModuleDescription = deploymentUnit.getAttachment(Attachments.EE_MODULE_DESCRIPTION);
		final WarMetaData metaData = deploymentUnit.getAttachment(WarMetaData.ATTACHMENT_KEY);
		if (metaData != null) {
			String serverNode = System.getProperty("jboss.node.name");
			String root = metaData.getJBossWebMetaData().getContextRoot();
			if (root == null) {
				root = eeModuleDescription.getModuleName();
			}
			Map<String, String> servlets = new HashMap<>();
			for (ServletMetaData curServlet : metaData.getWebMetaData().getServlets()) {
				servlets.put(curServlet.getServletName(), curServlet.getServletClass());
			}
			for (ServletMappingMetaData meta : metaData.getWebMetaData().getServletMappings()) {
				for (String url : meta.getUrlPatterns()) {
					if (url.startsWith("/")) {
						url = url.substring(1);
					}
					String alias = serverNode + "/" + root + "/" + url;
					ComponentRecord record = monitoringService.requireComponent(serverNode,
							eeModuleDescription.getApplicationName(), eeModuleDescription.getModuleName(),
							url);
					record.setComponentClass(servlets.get(meta.getServletName()));
					monitoringService.addAlias(alias, record);
				}
			}
		}

    	for (ComponentDescription componentDescription : eeModuleDescription.getComponentDescriptions()) {
			if (componentDescription instanceof EJBComponentDescription) {
				EJBComponentDescription ejb = (EJBComponentDescription) componentDescription;
				log.info("Register server-side interceptor for EJB " + ejb.getEJBName());
				ejb.getConfigurators().add(new ComponentConfigurator() {
					@Override
					public void configure(DeploymentPhaseContext context, ComponentDescription description,
							ComponentConfiguration configuration) throws DeploymentUnitProcessingException {
						MonitoringService service = MonitoringService.INSTANCE;
						
						for (ViewConfiguration view : configuration.getViews()) {
							if (view instanceof EJBViewConfiguration) {
								EJBViewConfiguration ejbView = (EJBViewConfiguration) view;
								switch(ejbView.getMethodIntf()) {
								case LOCAL:
									instrumentLocalEJBMethods(service, description, ejbView);
									break;
								case REMOTE:
									instrumentRemoteEJBMethods(service, description, ejbView);
									break;
								case MESSAGE_ENDPOINT:
									EJBMessageInterceptor interceptor = new EJBMessageInterceptor(service, description);
									view.addViewInterceptor(new ImmediateInterceptorFactory(interceptor), -1);
									break;
								case SERVICE_ENDPOINT:
									JAXWSDeployment jaxws = context.getDeploymentUnit()
											.getAttachment(WSAttachmentKeys.JAXWS_ENDPOINTS_KEY);
									instrumentWebServiceMethods(service, description, ejbView, jaxws);
									break;
								default:
									log.warn("Unsupported view type: " + ejbView.getMethodIntf());
									break;
								}
							}
						}
					}
				});
			} else if (componentDescription instanceof WSComponentDescription) {
				WSComponentDescription ws = (WSComponentDescription) componentDescription;
				log.info("Register server-side interceptor for web service " + ws.getComponentName());
				ws.getConfigurators().add(new ComponentConfigurator() {
					@Override
					public void configure(DeploymentPhaseContext context, ComponentDescription description,
							ComponentConfiguration configuration) throws DeploymentUnitProcessingException {
						MonitoringService service = MonitoringService.INSTANCE;
						JAXWSDeployment jaxws = context.getDeploymentUnit()
								.getAttachment(WSAttachmentKeys.JAXWS_ENDPOINTS_KEY);
						for (ViewConfiguration view : configuration.getViews()) {
							instrumentWebServiceMethods(service, description, view, jaxws);
						}
					}
				});
			}
		}

        log.info("Successfully instrumented deployment unit: " + deploymentUnit.getName());
    }

    @Override
    public void undeploy(DeploymentUnit context) {
    }

	private void instrumentLocalEJBMethods(MonitoringService monitoringService, ComponentDescription description,
			EJBViewConfiguration viewConfiguration) {
		for (Method curMethod : viewConfiguration.getProxyFactory().getCachedMethods()) {
			if (!curMethod.getDeclaringClass().equals(Object.class)) {
				if (Modifier.isPublic(curMethod.getModifiers())) {
					EJBLocalCallInterceptor interceptor = new EJBLocalCallInterceptor(monitoringService, description,
							viewConfiguration.getViewClass().getName(), curMethod);
					viewConfiguration.addViewInterceptor(curMethod, new ImmediateInterceptorFactory(interceptor), -1);
				}
			}
		}
	}

	private void instrumentRemoteEJBMethods(MonitoringService monitoringService, ComponentDescription description,
			EJBViewConfiguration viewConfiguration) {
		for (Method curMethod : viewConfiguration.getProxyFactory().getCachedMethods()) {
			if (!curMethod.getDeclaringClass().equals(Object.class)) {
				if (Modifier.isPublic(curMethod.getModifiers())) {
					EJBRemoteCallInterceptor interceptor = new EJBRemoteCallInterceptor(monitoringService, description,
							viewConfiguration.getViewClass().getName(), curMethod);
					viewConfiguration.addViewInterceptor(curMethod, new ImmediateInterceptorFactory(interceptor), -1);
				}
			}
		}
	}

	private void instrumentWebServiceMethods(MonitoringService monitoringService, ComponentDescription description,
			ViewConfiguration viewConfiguration, JAXWSDeployment jaxws) {
		List<AbstractEndpoint> allEndpoints = new ArrayList<>();
		allEndpoints.addAll(jaxws.getEjbEndpoints());
		allEndpoints.addAll(jaxws.getPojoEndpoints());

		// Find corresponding service endpoint
		AbstractEndpoint endpoint = null;
		for (AbstractEndpoint cur : allEndpoints) {
			if (cur.getComponentViewName().equals(viewConfiguration.getViewServiceName())) {
				endpoint = cur;
				break;
			}
		}
		if (endpoint != null) {
			// IMPORTANT: Use proxy factory to obtain
			// methods handles. Directly asking the view
			// type returns different method handles which
			// are not used by wildfly for interceptor
			// lookup.
			for (Method curMethod : viewConfiguration.getProxyFactory().getCachedMethods()) {
				if (!curMethod.getDeclaringClass().equals(Object.class)) {
					if (Modifier.isPublic(curMethod.getModifiers())) {
						if (endpoint instanceof POJOEndpoint) {
							WebServiceCallInterceptor interceptor = new WebServiceCallInterceptor(monitoringService,
									description, (POJOEndpoint) endpoint, curMethod);
							viewConfiguration.addViewInterceptor(curMethod,
									new ImmediateInterceptorFactory(interceptor), -1);
						} else if (endpoint instanceof EJBEndpoint) {
							EJBWebServiceCallInterceptor interceptor = new EJBWebServiceCallInterceptor(
									monitoringService, description, viewConfiguration.getViewClass().getName(),
									curMethod);
							viewConfiguration.addViewInterceptor(curMethod,
									new ImmediateInterceptorFactory(interceptor), -1);
						}
					}
				}
			}
		} else {
			log.warn("No endpoint found for web service: " + description.getComponentClassName());
		}
	};

}
