package tools.descartes.prisma.agent.wildfly.interceptors;

import java.util.Deque;

import org.jboss.as.ee.component.ComponentDescription;
import org.jboss.invocation.Interceptor;
import org.jboss.invocation.InterceptorContext;
import org.jboss.logging.Logger;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

public abstract class AbstractInterceptor implements Interceptor {

	protected final Logger log = Logger.getLogger(this.getClass());

	protected final MonitoringService monitoringService;
	protected final String serverNode;
	protected final ComponentRecord component;
	protected final String interfaceName;
	protected final OperationRecord operation;
	protected final boolean debug;
	private final boolean valueCharacterization;
	private final int[] parameters;

	public AbstractInterceptor(MonitoringService monitoringService, String assemblyName,
			ComponentDescription description,
			String interf, String protocol, String method) {
		this.interfaceName = interf;
		this.monitoringService = monitoringService;
		this.serverNode = System.getProperty("jboss.node.name");
		this.debug = log.isDebugEnabled();

		component = monitoringService.requireComponent(serverNode, description.getApplicationName(),
				description.getModuleName(),
				assemblyName);
		component.setComponentClass(description.getComponentClassName());
		operation = monitoringService.requireOperation(component, interf, protocol, method);

		// check whether we need to observe parameter values and set the fields
		// accordingly.
		parameters = operation.getObservedParameterIndeces();
		valueCharacterization = (parameters.length > 0);
	}

	protected InvocationRecord startInvocation(InterceptorContext ctx) {
		Deque<InvocationRecord> threadCallStack = monitoringService.getCallStack();
		// Push the current record onto the local thread stack

		InvocationRecord invocation;
		if (valueCharacterization) {
			// TODO: currently only a single parameter can be observed
			invocation = operation.startInvocation(ctx.getParameters()[parameters[0]]);
		} else {
			invocation = operation.startInvocation();
		}
		threadCallStack.push(invocation);
		return invocation;
	}

	protected void finishInvocation(long startTimeInNs) {
		Deque<InvocationRecord> threadCallStack = monitoringService.getCallStack();

		InvocationRecord invocation = threadCallStack.pop();
		InvocationRecord localOutgoingCall = null;

		if (!threadCallStack.isEmpty()) {
			InvocationRecord caller = threadCallStack.peek();
			// this is a local call in the same thread
			// also add a outgoing call to the callee component
			if (debug) {
				log.debug("Outgoing invocation (" + caller.getOperation().getInterfaceName() + ", "
						+ caller.getOperation().getOperationName() + ") -> ("
						+ interfaceName + ", " + operation + ")");
			}
			OperationRecord callerOrec = caller.requireEndpoint(component, interfaceName, "java",
					operation.getOperationName());
			localOutgoingCall = callerOrec.getInvocations()[invocation.getValueBucket()];
		}

		// Do the measurement at the latest possible instant in time to include
		// overheads of the interceptor within the measurement.
		long executionTime = System.nanoTime() - startTimeInNs;
		invocation.finishInvocation(executionTime, 0, 0);
		if (localOutgoingCall != null) {
			localOutgoingCall.finishInvocation(executionTime, 0, 0);
		}
		if (debug) {
			log.debug("Incoming invocation (" + operation.getComponent().getComponentName() + ", "
					+ operation.getOperationName() + ", " + executionTime + "ms)");
		}
	}

}
