package tools.descartes.prisma.agent.wildfly.interceptors;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.lang.reflect.Field;
import java.util.Deque;

import org.apache.commons.lang.mutable.MutableLong;
import org.jboss.ejb.client.EJBClientInterceptor;
import org.jboss.ejb.client.EJBClientInvocationContext;
import org.jboss.ejb.client.EJBReceiverInvocationContext;
import org.jboss.logging.Logger;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

public class EJBOutgoingCallInterceptor implements EJBClientInterceptor {
	
	private static final Logger log = Logger.getLogger(EJBOutgoingCallInterceptor.class);
	
	private final ThreadMXBean threadBean;
	
	private Field receiverInvocationContextField = null;

	private final MonitoringService monitoringService;

	public EJBOutgoingCallInterceptor(MonitoringService monitoringService) {
		this.monitoringService = monitoringService;
		try {
			receiverInvocationContextField = EJBClientInvocationContext.class
					.getDeclaredField("receiverInvocationContext");
			receiverInvocationContextField.setAccessible(true);
		} catch (Exception e) {
			log.error("Error getting receiver invocation field.", e);
		}
		threadBean = ManagementFactory.getThreadMXBean();
		if (threadBean.isCurrentThreadCpuTimeSupported()) {
			threadBean.setThreadCpuTimeEnabled(true);
		}
	}
	
	private static final ThreadLocal<MutableLong> invocationTime = new ThreadLocal<MutableLong>() {
		protected MutableLong initialValue() {
			return new MutableLong();
		}
	};
	
	private static final ThreadLocal<MutableLong> startCpuTime = new ThreadLocal<MutableLong>() {
		protected MutableLong initialValue() {
			return new MutableLong();
		}
	};

	@Override
	public void handleInvocation(EJBClientInvocationContext ctx) throws Exception {
		Deque<InvocationRecord> threadCallStack = monitoringService.getCallStack();
		InvocationRecord currentInvocation = threadCallStack.peek();
		if (currentInvocation != null) {
			String interf = ctx.getViewClass().getName();
			String operation = ctx.getInvokedMethod().getName();

			String nodeName = ((EJBReceiverInvocationContext) receiverInvocationContextField.get(ctx)).getNodeName();
			ComponentRecord component = monitoringService.requireComponent(nodeName,
					ctx.getLocator().getAppName(), ctx.getLocator().getModuleName(), ctx.getLocator().getBeanName());

			OperationRecord orec = currentInvocation.requireEndpoint(component, interf, "rmi",
					operation);
			InvocationRecord outgoingInvocation;
			if (orec.areParametersMonitored()) {
				outgoingInvocation = orec.startInvocation(ctx.getParameters()[orec.getObservedParameterIndeces()[0]]);
			} else {
				outgoingInvocation = orec.startInvocation();
			}
			
			threadCallStack.push(outgoingInvocation);
			
			invocationTime.get().setValue(System.nanoTime());
//			startCpuTime.get().setValue(threadBean.getCurrentThreadCpuTime());
		}

		ctx.sendRequest();
	}

	@Override
	public Object handleInvocationResult(EJBClientInvocationContext ctx) throws Exception {
		try {
			return ctx.getResult();
		} finally {
			long endTime = System.nanoTime();
//			long endCpuTime = threadBean.getCurrentThreadCpuTime();
			Deque<InvocationRecord> threadCallStack = monitoringService.getCallStack();
			if (!threadCallStack.isEmpty()) {
				long executionTime = endTime - invocationTime.get().longValue();
//				long cpuTime = endCpuTime - startCpuTime.get().longValue();
				InvocationRecord invocation = threadCallStack.pop();
				invocation.finishInvocation(executionTime, 0, 0);
					
				if (log.isTraceEnabled()) {
					OperationRecord caller = threadCallStack.peek().getOperation();
					log.trace("Outgoing invocation (" + caller.getInterfaceName() + ", "
							+ caller.getOperationName() + ") -> ("
							+ invocation.getOperation().getInterfaceName() + ", "
							+ invocation.getOperation().getOperationName() + ", " + executionTime + "ms)");
				}
			}
		}		
	}
}
