package tools.descartes.prisma.agent.wildfly.service;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import org.jboss.logging.Logger;
import org.jboss.msc.service.Service;
import org.jboss.msc.service.ServiceName;
import org.jboss.msc.service.StartContext;
import org.jboss.msc.service.StartException;
import org.jboss.msc.service.StopContext;
import org.jboss.msc.value.InjectedValue;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.parameters.ValueCharacterization;


public class MonitoringService implements Service<MonitoringService> {
	
	private static final Logger log = Logger.getLogger(MonitoringService.class);
	
	public static final ServiceName SERVICE_NAME = ServiceName.JBOSS.append("modelextraction", "monitoring");
	
	public static final ServiceName FILTER_NAME = SERVICE_NAME.append("filter");
	
	public static final ServiceName PARAM_CHARACTERIZATION_NAME = SERVICE_NAME.append("filter");
	
	public static final MonitoringService INSTANCE = new MonitoringService();
	
	private static final long WRITE_INTERVAL = 60000;
	
	private final ConcurrentHashMap<Long, Long> threads = new ConcurrentHashMap<Long, Long>();
	
	private List<InjectedValue<ParameterCharacterizationService>> paramCharacterizations = new CopyOnWriteArrayList<>();

	private Map<String, Map<String, ValueCharacterization>> valueCharacterizations = new HashMap<>();

	private final ConcurrentHashMap<String, OperationRecord> monitoredEndpoints = new ConcurrentHashMap<>();

	private final ConcurrentHashMap<String, ComponentRecord> components = new ConcurrentHashMap<>();

	private static final ThreadLocal<Deque<InvocationRecord>> callStack = new ThreadLocal<Deque<InvocationRecord>>() {
		protected Deque<InvocationRecord> initialValue() {
			return new ArrayDeque<InvocationRecord>(10);
		};
	};

	private Thread OUTPUT = new Thread() {
		public void run() {
			try {
				while(true) {
					long currentTime = System.currentTimeMillis();
					// always do it at event intervals
					long nextTime = (currentTime / WRITE_INTERVAL + 1) * WRITE_INTERVAL;
					Thread.sleep(nextTime - currentTime);
					
					try(final PrintWriter out = new PrintWriter(new FileWriter(new File("./operations.csv"), true))) {
						writeMonitoringDataToDisk(out, nextTime);
					} catch(Exception ex) {
						log.error("Error writing monitoring data.", ex);
					}
				}
			} catch(InterruptedException ex) {
				interrupted();
			}
		};
	};

	@Override
	public MonitoringService getValue() throws IllegalStateException, IllegalArgumentException {
		return this;
	}

	@Override
	public void start(StartContext ctx) throws StartException {
		for (InjectedValue<ParameterCharacterizationService> param : paramCharacterizations) {
			addValueCharacterization(param.getValue().createValueCharacterization());
		}
		System.err.println("MonitoringService started");		
		OUTPUT.start();	
	}

	@Override
	public void stop(StopContext arg0) {
		OUTPUT.interrupt();
	}
	
	private void writeMonitoringDataToDisk(PrintWriter out, long timestamp) {
		for (OperationRecord curEndpoint : monitoredEndpoints.values()) {
			for (InvocationRecord curInvocation : curEndpoint.getInvocations()) {
				out.printf("%d, %s, , %d, %d, %s\n", timestamp, curInvocation.toString(),
						curInvocation.getInvocationCount(), curInvocation.getExecutionTime(),
						curInvocation.getAdditionalInformation());

				for (OperationRecord curOutgoingEndpoint : curInvocation.getCalledEndpoints()) {
					for (InvocationRecord curOutgoingInvocation : curOutgoingEndpoint.getInvocations()) {
						out.printf("%d, %s, %s, %d, %d, %s\n", timestamp, curInvocation.toString(),
								curOutgoingInvocation.toString(), curOutgoingInvocation.getInvocationCount(),
								curOutgoingInvocation.getExecutionTime(),
								curOutgoingInvocation.getAdditionalInformation());
					}
				}
			}
		}
	}
	
	public long getLastCpuTime(long threadId) {
		Long val = threads.get(threadId);
		return (val == null) ? 0 : val;
	}
	
	public void setLastCpuTime(long threadId, long cpuTimeInNanos) {
		threads.put(threadId, cpuTimeInNanos);
	}
	
	public List<InjectedValue<ParameterCharacterizationService>> getParameterCharacterizations() {
		return paramCharacterizations;
	}

	public ComponentRecord addAlias(String alias, ComponentRecord record) {
		return components.put(alias, record);
	}

	public void addValueCharacterization(ValueCharacterization charac) {
		System.err.println("Value characterization added");
		Map<String, ValueCharacterization> ops = valueCharacterizations.get(charac.getInterfaceName());
		if (ops == null) {
			ops = new HashMap<>();
			valueCharacterizations.put(charac.getInterfaceName(), ops);
		}
		ops.put(charac.getOperationName(), charac);
	}

	public OperationRecord createOperationRecord(ComponentRecord component, String interf, String protocol,
			String operation) {
		return createOperationRecord(null, component, interf, protocol, operation);
	}

	public OperationRecord createOperationRecord(InvocationRecord caller, ComponentRecord component, String interf,
			String protocol, String operation) {
		ValueCharacterization valueCharac = null;
		Map<String, ValueCharacterization> charac = valueCharacterizations.get(interf);
		if (charac != null) {
			valueCharac = charac.get(operation);
		}
		return new OperationRecord(this, caller, component, interf, protocol, operation, valueCharac);
	}

	public OperationRecord requireOperation(ComponentRecord component, String interf, String protocol,
			String operation) {
		String endpointIdentifier = OperationRecord.getEndpointIdentification(component, interf, protocol, operation);
		OperationRecord record = monitoredEndpoints.get(endpointIdentifier);
		if (record == null) {
			record = createOperationRecord(component, interf, protocol, operation);
			if (monitoredEndpoints.putIfAbsent(endpointIdentifier, record) != null) {
				// a different thread has changed it in the mean time
				record = monitoredEndpoints.get(endpointIdentifier);
			} else {
				ModelExtractionService.INSTANCE.newOperationRecord(record);
			}
		}
		return record;
	}

	public ComponentRecord getComponent(String identifier) {
		return components.get(identifier);
	}

	public ComponentRecord requireComponent(String serverNode, String application, String module, String component) {
		String componentIdentifier = ComponentRecord.getComponentIdentifier(serverNode, application, module, component);
		ComponentRecord record = components.get(componentIdentifier);
		if (record == null) {
			record = new ComponentRecord(serverNode, application, module, component);
			if (components.putIfAbsent(componentIdentifier, record) != null) {
				// a different thread has changed it in the mean time
				record = components.get(componentIdentifier);
			} else {
				ModelExtractionService.INSTANCE.newComponentRecord(record);
			}
		}
		return record;
	}

	public Collection<OperationRecord> getEndpoints() {
		return monitoredEndpoints.values();
	}

	public Deque<InvocationRecord> getCallStack() {
		return callStack.get();
	}

}
