package tools.descartes.prisma.core;

import static org.fest.assertions.api.Assertions.assertThat;

import java.net.InetAddress;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.junit.Before;
import org.junit.Test;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.Interface;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Repository;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter.ChangeStatus;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.scopes.ApplicationScope;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;
import tools.descartes.prisma.model.skeleton.SkeletonFactory;

public class ModelSkeletonTest {
	private String name;
	private ModelRepository repository;
	private MessageBus messageBus;
	private ApplicationScope scope;
	private AgentController controller;

	@Before
	public void setUp() throws Exception {
		name = "test";
		messageBus = new TestMessageBus();
		repository = new TestModelRepository(messageBus);
		repository.getSystemScope().create();
		this.controller = new AgentController(repository, messageBus,
				ApplicationScope.APPLICATION_SCOPE_TYPE + "/" + name,
				InetAddress.getLocalHost().toString());
		this.controller.start().get();
		this.scope = (ApplicationScope) this.controller.getContainerScope();
	}

	@Test
	public void testAddNewInterface() {
		ModelSkeleton skeleton = SkeletonFactory.eINSTANCE.createModelSkeleton();
		Repository repository = RepositoryFactory.eINSTANCE.createRepository();
		repository.getInterfaces().add(createInterface("test"));
		skeleton.setRepository(repository);

		controller.apply(skeleton);

		assertThat(scope.getComponentRepository().getInterfaces()).hasSize(1);
		assertInterface(scope.getComponentRepository().getInterfaces().get(0), "test");
		assertUnchanged(skeleton);
	}

	@Test
	public void testAddSignaturesToInterface() {
		ModelSkeleton skeleton = SkeletonFactory.eINSTANCE.createModelSkeleton();
		Repository repository = RepositoryFactory.eINSTANCE.createRepository();
		Interface interf = createInterface("test", "call1");
		repository.getInterfaces().add(interf);
		skeleton.setRepository(repository);

		controller.apply(skeleton);

		assertThat(scope.getComponentRepository().getInterfaces()).hasSize(1);
		assertInterface(scope.getComponentRepository().getInterfaces().get(0), "test", "call1");
		assertUnchanged(skeleton);

		Signature signature2 = RepositoryFactory.eINSTANCE.createSignature();
		signature2.setName("call2");
		interf.getSignatures().add(signature2);

		controller.apply(skeleton);

		assertInterface(scope.getComponentRepository().getInterfaces().get(0), "test", "call1", "call2");
		assertUnchanged(skeleton);
	}

	@Test
	public void testAddExistingInterface() {
		// There is already an interface with the same name in the repository
		// Simulate it by manually adding it to the scope.
		scope.getComponentRepository().getInterfaces().add(createInterface("test", "call1"));

		// Create model skeleton. The interface in the model skeleton should be
		// matched with the one in the repository.
		ModelSkeleton skeleton = SkeletonFactory.eINSTANCE.createModelSkeleton();
		Repository repository = RepositoryFactory.eINSTANCE.createRepository();
		repository.getInterfaces().add(createInterface("test", "call2"));
		skeleton.setRepository(repository);

		controller.apply(skeleton);

		// There should be only one interface with name "test"
		assertThat(scope.getComponentRepository().getInterfaces()).hasSize(1);
		assertInterface(scope.getComponentRepository().getInterfaces().get(0), "test", "call1", "call2");
		assertUnchanged(skeleton);
	}

	private Interface createInterface(String name, String... signatures) {
		Interface interf = RepositoryFactory.eINSTANCE.createInterface();
		interf.setName(name);
		for (String curSignature : signatures) {
			Signature s = RepositoryFactory.eINSTANCE.createSignature();
			s.setName(curSignature);
			interf.getSignatures().add(s);
		}
		return interf;
	}

	private Interface assertInterface(Interface interf, String name, String... signatures) {
		assertThat(interf.getName()).isEqualTo(name);
		assertThat(interf.getSignatures()).hasSameSizeAs(signatures);
		for (int i = 0; i < signatures.length; i++) {
			assertThat(interf.getSignatures().get(i).getName()).isEqualTo(signatures[i]);
		}
		return interf;
	}

	private void assertUnchanged(ModelSkeleton skeleton) {
		TreeIterator<EObject> iter = skeleton.eAllContents();
		while (iter.hasNext()) {
			ModelSkeletonAdapter adapter = controller.adapt(iter.next(), ModelSkeletonAdapter.class);
			assertThat(adapter.getChangeStatus()).isEqualTo(ChangeStatus.UNCHANGED);
		}
	}

}
