package tools.descartes.prisma.core.adapter;

import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.exceptions.ConcurrentTransactionException;
import tools.descartes.prisma.core.merge.MergerFactory;
import tools.descartes.prisma.core.merge.ObjectMerger;

public class ModelSkeletonAdapter extends AdapterImpl {
	
	private static final Log log = LogFactory.getLog(ModelSkeletonAdapter.class);

	public static enum ChangeStatus {
		NEW, UNCHANGED, CHANGED
	}
	
	protected static class TransactionContext {
		public Transaction currentTransaction = null;
	}


	// When the adapter is created for an element, it is always 
	// considered to be new. This is checked later
	private ChangeStatus status = ChangeStatus.NEW;
	private CDOObject repositoryObjectReadOnly;
	private CDOObject changedRepositoryObject = null;
	private final TransactionContext transactionStatus;
	private final List<Notification> changes = new LinkedList<>();
	private final ModelRepository repository;
	private final AgentController agent;
	private final ObjectMerger merger;
	private boolean isRoot = false;
	// The class of the observed object
	private EClass objectType;
	// This is only used for log messages to uniquely identify the object
	private String objectURI = "";

	public ModelSkeletonAdapter(ModelRepository repository, AgentController agent, TransactionContext context,
			EClass type) {
		this.repository = repository;
		this.transactionStatus = context;
		this.agent = agent;
		this.objectType = type;
		this.merger = MergerFactory.createMerger(agent, type);
	}
	
	@Override
	public boolean isAdapterForType(Object type) {
		return getClass().equals(type);
	}

	@Override
	public void notifyChanged(Notification msg) {
		switch (msg.getEventType()) {
		case Notification.ADD:
		case Notification.ADD_MANY:
		case Notification.REMOVE:
		case Notification.REMOVE_MANY:
		case Notification.SET:
		case Notification.UNSET:
			if (log.isDebugEnabled()) {
				log.debug("Received change notification (object=" + objectURI + ", feature="
						+ msg.getFeatureID(objectType.getClass()) + ", event=" + msg.getEventType() + ")");
			}
			changes.add(msg);
			status = ChangeStatus.CHANGED;
			break;
		default:
			// ignore event
		}
	}
	
	public ChangeStatus getChangeStatus() {
		return status;
	}

	public CDOObject getRepositoryObject() {
		return this.repositoryObjectReadOnly;
	}

	public void setRepositoryObject(CDOObject repositoryObject) {
		if (log.isDebugEnabled()) {
			log.debug("Set repository object (object=" + objectURI + ", cdoid=" + repositoryObject.cdoID()
					+ ")");
		}
		this.repositoryObjectReadOnly = repositoryObject;
		changes.clear();
		status = ChangeStatus.UNCHANGED;
	}

	public void startTransaction(Transaction transaction) {
		if (log.isDebugEnabled()) {
			log.debug("Start transaction (object=" + objectURI + ")");
		}
		if ((transactionStatus.currentTransaction != null) && (transactionStatus.currentTransaction != transaction)) {
			throw new ConcurrentTransactionException();
		}
		this.transactionStatus.currentTransaction = transaction;
	}

	public void updateRepositoryObject(CDOObject repositoryObject) {
		if (log.isDebugEnabled()) {
			log.debug("Set transient object (object=" + objectURI + ", cdoid=" + repositoryObject.cdoID() + ")");
		}
		checkTransaction();

		this.changedRepositoryObject = transactionStatus.currentTransaction.attach(repositoryObject);
	}

	public void synchronizeRoot(CDOObject root) {
		if (log.isDebugEnabled()) {
			log.debug("Synchronize root (object=" + objectURI + ")");
		}
		checkTransaction();

		isRoot = true;
		updateRepositoryObject(root);

		synchronizeTree();
	}

	protected void synchronizeTree() {
		if (log.isDebugEnabled()) {
			log.debug("Synchronize tree (object=" + objectURI + ")");
		}
		EObject changedObject = getCastedTarget();
		merger.mergeTree(transactionStatus.currentTransaction, getTransientRepositoryObject(), changedObject, null);
		// Synchronize all contained children.
		for (EObject child : changedObject.eContents()) {
			ModelSkeletonAdapter childAdapter = agent.adapt(child, ModelSkeletonAdapter.class);
			childAdapter.synchronizeTree();
		}
	}

	private void checkTransaction() {
		if (transactionStatus.currentTransaction == null) {
			throw new IllegalStateException("No transaction is currently running.");
		}
	}

	public CDOObject getTransientRepositoryObject() {
		if (transactionStatus.currentTransaction == null) {
			return repositoryObjectReadOnly;
		} else {
			if (changedRepositoryObject == null) {
				if (repositoryObjectReadOnly == null) {
					// this case usually happens if the skeleton is invalid
					// (e.g., dangling references)
					return null;
				} else {
					changedRepositoryObject = transactionStatus.currentTransaction
						.attach(repositoryObjectReadOnly);
				}
			}
			return changedRepositoryObject;
		}
	}

	public void synchronizeContents() {
		if (log.isDebugEnabled()) {
			log.debug("Synchronize contents (object=" + objectURI + ")");
		}
		checkTransaction();

		EObject changedObject = getCastedTarget();
		// The root elements of the model skeleton are just containers that are
		// not sync'ed
		if (!isRoot) {
			merger.mergeContents(transactionStatus.currentTransaction, getTransientRepositoryObject(), changedObject,
					null);
		}
		// Synchronize features of contained children.
		for (EObject child : changedObject.eContents()) {
			ModelSkeletonAdapter childAdapter = agent.adapt(child, ModelSkeletonAdapter.class);
			childAdapter.synchronizeContents();
		}
	}

	public ModelSkeletonDelta commitFinished(boolean success) {
		if (log.isDebugEnabled()) {
			log.debug("Transaction commit finished (object=" + objectURI + ", success=" + success + ")");
		}
		if (transactionStatus.currentTransaction == null) {
			// No transaction is currently running
			return new ModelSkeletonDelta();
		}

		ModelSkeletonDelta delta = new ModelSkeletonDelta();
		commitFinishedWithDelta(success, delta);
		transactionStatus.currentTransaction = null;
		return delta;
	}

	private void commitFinishedWithDelta(boolean success, ModelSkeletonDelta delta) {
		// Synchronize features of contained children.
		EObject target = getCastedTarget();
		for (EObject child : target.eContents()) {
			ModelSkeletonAdapter childAdapter = agent.adapt(child, ModelSkeletonAdapter.class);
			childAdapter.commitFinishedWithDelta(success, delta);
		}

		if (success) {
			changes.clear();
			if (repositoryObjectReadOnly == null) {
				repositoryObjectReadOnly = transactionStatus.currentTransaction
.detach(changedRepositoryObject);
			}
			switch (status) {
			case NEW:
				delta.addNewObject(repositoryObjectReadOnly);
				break;
			case CHANGED:
				delta.addChangedObject(repositoryObjectReadOnly);
				break;
			case UNCHANGED:
				break;
			}
			status = ChangeStatus.UNCHANGED;
		}

		changedRepositoryObject = null;
	}

	@Override
	public void setTarget(Notifier newTarget) {
		super.setTarget(newTarget);
		objectURI = EcoreUtil.getURI((EObject) newTarget).toString();
	}

	protected EObject getCastedTarget() {
		return (EObject) getTarget();
	}

}
