package tools.descartes.prisma.core.adapter;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import edu.kit.ipd.descartes.core.CorePackage;
import edu.kit.ipd.descartes.identifier.IdentifierPackage;
import edu.kit.ipd.descartes.mm.applicationlevel.functions.FunctionsPackage;
import edu.kit.ipd.descartes.mm.applicationlevel.parameterdependencies.ParameterdependenciesPackage;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryPackage;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorPackage;
import edu.kit.ipd.descartes.mm.resourceconfiguration.ResourceconfigurationPackage;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapePackage;
import edu.kit.ipd.descartes.mm.resourcetype.ResourcetypePackage;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfilePackage;

public class ObjectMatcher {

	private static class StructuralFeatureMatcher {

		private final EStructuralFeature[] identityFeatures;

		public StructuralFeatureMatcher(EStructuralFeature[] features) {
			this.identityFeatures = features;
		}

		public boolean matches(EObject o1, EObject o2) {
			for (EStructuralFeature curFeature : identityFeatures) {
				boolean isSet1 = o1.eIsSet(curFeature);
				boolean isSet2 = o2.eIsSet(curFeature);
				if (!isSet1 || !isSet2) {
					if (!isSet1 && !isSet2) {
						continue;
					}
					return false;
				}
				Object v1 = o1.eGet(curFeature);
				Object v2 = o2.eGet(curFeature);
				if (v1 == null || v2 == null) {
					// Identity features should never be null
					return false;
				}
				// Attributes can be compared directly, while for references
				// we need to delegate the matching again to a 
				// StructuralFeatureMatcher
				if (!curFeature.isMany()) {
					if (curFeature instanceof EAttribute) {
						if (!v1.equals(v2)) {
							return false;
						}
					} else {
						if (!ObjectMatcher.matches((EObject) v1, (EObject) v2)) {
							return false;
						}
					}
				} else {
					List<?> l1 = (List<?>) v1;
					List<?> l2 = (List<?>) v2;
					if (l1.size() != l2.size()) {
						return false;
					} else {
						if (curFeature instanceof EAttribute) {
							for (int i = 0; i < l1.size(); i++) {
								if (!l1.get(i).equals(l2.get(i))) {
									return false;
								}
							}
						} else {
							for (int i = 0; i < l1.size(); i++) {
								if (!ObjectMatcher.matches((EObject) l1.get(i), (EObject) l2.get(i))) {
									return false;
								}
							}
						}
					}
				}
			}
			return true;
		}

	}

	/*
	 * This is a map containing feature matchers for different classes used to
	 * customized the matching behavior.
	 */
	private static Map<EClass, StructuralFeatureMatcher> featureMatchers = new HashMap<>();

	/*
	 * This is the fallback matcher used for all classes not contained in
	 * featureMatchers map
	 */
	private static StructuralFeatureMatcher defaultMatcher = features(IdentifierPackage.Literals.IDENTIFIER__ID);


	static {
		// Repository package: This package is on a type level, i.e., the ID
		// attribute
		// is not sufficient for matching. For instance an interface with the
		// same
		// name should be matched even if they have different ID values.
		featureMatchers.put(RepositoryPackage.Literals.PRIMITIVE_DATA_TYPE, features(
				CorePackage.Literals.NAMED_ELEMENT__NAME,
				RepositoryPackage.Literals.PRIMITIVE_DATA_TYPE__TYPE));
		featureMatchers.put(RepositoryPackage.Literals.COLLECTION_DATA_TYPE, features(
				CorePackage.Literals.NAMED_ELEMENT__NAME,
				RepositoryPackage.Literals.COLLECTION_DATA_TYPE__INNER_TYPE));
		featureMatchers.put(RepositoryPackage.Literals.COMPOSITE_DATA_TYPE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.INNER_DECLARATION, features(
				CorePackage.Literals.NAMED_ELEMENT__NAME,
				RepositoryPackage.Literals.INNER_DECLARATION__DATA_TYPE));
		featureMatchers.put(RepositoryPackage.Literals.INTERFACE, features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.SIGNATURE, features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.INPUT_PARAMETER,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME, RepositoryPackage.Literals.PARAMETER__DATA_TYPE));
		featureMatchers.put(RepositoryPackage.Literals.OUTPUT_PARAMETER,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME, RepositoryPackage.Literals.PARAMETER__DATA_TYPE));
		featureMatchers.put(RepositoryPackage.Literals.INPUT_OUTPUT_PARAMETER,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME, RepositoryPackage.Literals.PARAMETER__DATA_TYPE));
		featureMatchers.put(RepositoryPackage.Literals.BASIC_COMPONENT,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.SEMAPHORE, features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.COMPOSITE_COMPONENT,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.SUB_SYSTEM, features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(RepositoryPackage.Literals.INTERFACE_PROVIDING_ROLE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
				RepositoryPackage.Literals.INTERFACE_PROVIDING_ROLE__INTERFACE));
		featureMatchers.put(RepositoryPackage.Literals.INTERFACE_REQUIRING_ROLE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
				RepositoryPackage.Literals.INTERFACE_REQUIRING_ROLE__INTERFACE));
		featureMatchers.put(RepositoryPackage.Literals.ASSEMBLY_CONTEXT,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
				RepositoryPackage.Literals.ASSEMBLY_CONTEXT__ENCAPSULATED_COMPONENT));
		featureMatchers.put(RepositoryPackage.Literals.ASSEMBLY_CONNECTOR,
				features(RepositoryPackage.Literals.ASSEMBLY_CONNECTOR__PROVIDING_ASSEMBLY_CONTEXT,
						RepositoryPackage.Literals.ASSEMBLY_CONNECTOR__REQUIRING_ASSEMBLY_CONTEXT,
						RepositoryPackage.Literals.ASSEMBLY_CONNECTOR__INTERFACE_PROVIDING_ROLE,
						RepositoryPackage.Literals.ASSEMBLY_CONNECTOR__INTERFACE_REQUIRING_ROLE));
		featureMatchers.put(RepositoryPackage.Literals.PROVIDING_DELEGATION_CONNECTOR,
				features(RepositoryPackage.Literals.DELEGATION_CONNECTOR__ASSEMBLY_CONTEXT,
						RepositoryPackage.Literals.PROVIDING_DELEGATION_CONNECTOR__INNER_INTERFACE_PROVIDING_ROLE,
						RepositoryPackage.Literals.PROVIDING_DELEGATION_CONNECTOR__OUTER_INTERFACE_PROVIDING_ROLE));
		featureMatchers.put(RepositoryPackage.Literals.REQUIRING_DELEGATION_CONNECTOR,
				features(RepositoryPackage.Literals.DELEGATION_CONNECTOR__ASSEMBLY_CONTEXT,
						RepositoryPackage.Literals.REQUIRING_DELEGATION_CONNECTOR__INNER_INTERFACE_REQUIRING_ROLE,
						RepositoryPackage.Literals.REQUIRING_DELEGATION_CONNECTOR__OUTER_INTERFACE_REQUIRING_ROLE));
		// Servicebehavior package:
		featureMatchers.put(ServicebehaviorPackage.Literals.BLACK_BOX_BEHAVIOR,
				features(ServicebehaviorPackage.Literals.SERVICE_BEHAVIOR_ABSTRACTION__DESCRIBED_SIGNATURE,
						ServicebehaviorPackage.Literals.SERVICE_BEHAVIOR_ABSTRACTION__INTERFACE_PROVIDING_ROLE));
		featureMatchers.put(ServicebehaviorPackage.Literals.RESPONSE_TIME,
				features(ParameterdependenciesPackage.Literals.INFLUENCABLE_MODEL_VARIABLE__NAME));
		featureMatchers.put(ServicebehaviorPackage.Literals.COARSE_GRAINED_BEHAVIOR,
				features(ServicebehaviorPackage.Literals.SERVICE_BEHAVIOR_ABSTRACTION__DESCRIBED_SIGNATURE,
						ServicebehaviorPackage.Literals.SERVICE_BEHAVIOR_ABSTRACTION__INTERFACE_PROVIDING_ROLE));
		featureMatchers.put(ServicebehaviorPackage.Literals.RESOURCE_DEMAND,
				features(ParameterdependenciesPackage.Literals.INFLUENCABLE_MODEL_VARIABLE__NAME,
						ServicebehaviorPackage.Literals.RESOURCE_DEMAND__RESOURCE_TYPE));
		featureMatchers.put(ServicebehaviorPackage.Literals.EXTERNAL_CALL_FREQUENCY,
				features(ServicebehaviorPackage.Literals.EXTERNAL_CALL_FREQUENCY__FREQUENCY,
						ServicebehaviorPackage.Literals.EXTERNAL_CALL_FREQUENCY__EXTERNAL_CALL));
		featureMatchers.put(ServicebehaviorPackage.Literals.CALL_FREQUENCY,
				features(ParameterdependenciesPackage.Literals.INFLUENCABLE_MODEL_VARIABLE__NAME));
		featureMatchers.put(ServicebehaviorPackage.Literals.EXTERNAL_CALL,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
						ServicebehaviorPackage.Literals.EXTERNAL_CALL__INTERFACE_REQUIRING_ROLE,
						ServicebehaviorPackage.Literals.EXTERNAL_CALL__SIGNATURE));
		featureMatchers.put(ServicebehaviorPackage.Literals.FINE_GRAINED_BEHAVIOR,
				features(ServicebehaviorPackage.Literals.SERVICE_BEHAVIOR_ABSTRACTION__DESCRIBED_SIGNATURE,
						ServicebehaviorPackage.Literals.SERVICE_BEHAVIOR_ABSTRACTION__INTERFACE_PROVIDING_ROLE));
		featureMatchers.put(ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR, features());
		featureMatchers.put(ServicebehaviorPackage.Literals.ACQUIRE_ACTION,
				features(ServicebehaviorPackage.Literals.ACQUIRE_ACTION__SEMAPHORE));
		featureMatchers.put(ServicebehaviorPackage.Literals.RELEASE_ACTION,
				features(ServicebehaviorPackage.Literals.RELEASE_ACTION__SEMAPHORE));
		featureMatchers.put(ServicebehaviorPackage.Literals.INTERNAL_ACTION,
				features(ServicebehaviorPackage.Literals.INTERNAL_ACTION__RESOURCE_DEMAND));
		featureMatchers.put(ServicebehaviorPackage.Literals.LOOP_ACTION,
				features(ServicebehaviorPackage.Literals.LOOP_ACTION__LOOP_ITERATION_COUNT));
		featureMatchers.put(ServicebehaviorPackage.Literals.FORK_ACTION, features());
		featureMatchers.put(ServicebehaviorPackage.Literals.BRANCH_ACTION,
				features(ServicebehaviorPackage.Literals.BRANCH_ACTION__BRANCH_PROBABILITIES));
		
		// UsageProfile package:
		featureMatchers.put(UsageProfilePackage.Literals.USAGE_SCENARIO, 
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
						UsageProfilePackage.Literals.USAGE_SCENARIO__WORKLOAD_TYPE));
		featureMatchers.put(UsageProfilePackage.Literals.CLOSED_WORKLOAD_TYPE, features());
		featureMatchers.put(UsageProfilePackage.Literals.OPEN_WORKLOAD_TYPE, features());
		featureMatchers.put(UsageProfilePackage.Literals.SCENARIO_BEHAVIOR, features());
		featureMatchers.put(UsageProfilePackage.Literals.BRANCH_USER_ACTION, features());
		featureMatchers.put(UsageProfilePackage.Literals.LOOP_USER_ACTION, features());
		featureMatchers.put(UsageProfilePackage.Literals.DELAY_USER_ACTION, features());
		featureMatchers.put(UsageProfilePackage.Literals.SYSTEM_CALL_USER_ACTION,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
						UsageProfilePackage.Literals.SYSTEM_CALL_USER_ACTION__PROVIDING_ROLE,
						UsageProfilePackage.Literals.SYSTEM_CALL_USER_ACTION__SIGNATURE));
		featureMatchers.put(UsageProfilePackage.Literals.CALL_PARAMETER_SETTING,
				features(UsageProfilePackage.Literals.CALL_PARAMETER_SETTING__NAME,
						UsageProfilePackage.Literals.CALL_PARAMETER_SETTING__CHARACTERIZATION));

		// Random Variable
		featureMatchers.put(FunctionsPackage.Literals.RANDOM_VARIABLE, features());
		featureMatchers.put(FunctionsPackage.Literals.PROBABILITY_MASS_FUNCTION, features());
		featureMatchers.put(FunctionsPackage.Literals.ENUM_SAMPLE_LIST, features());
		featureMatchers.put(FunctionsPackage.Literals.ENUM_SAMPLE, features());
		featureMatchers.put(FunctionsPackage.Literals.INT_SAMPLE_LIST, features());
		featureMatchers.put(FunctionsPackage.Literals.INT_SAMPLE, features());
		featureMatchers.put(FunctionsPackage.Literals.EXPONENTIAL_DISTRIBUTION, features());
		featureMatchers.put(FunctionsPackage.Literals.DOUBLE_LITERAL, features());

		// Resourcelandscape
		featureMatchers.put(ResourcelandscapePackage.Literals.DATA_CENTER,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(ResourcelandscapePackage.Literals.COMPUTE_NODE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(ResourcelandscapePackage.Literals.STORAGE_NODE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(ResourcelandscapePackage.Literals.RUNTIME_ENVIRONMENT,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
						ResourcelandscapePackage.Literals.RUNTIME_ENVIRONMENT__OF_CLASS));

		// Resourceconfiguration
		featureMatchers.put(ResourceconfigurationPackage.Literals.PROCESSING_RESOURCE_SPECIFICATION,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME,
						ResourceconfigurationPackage.Literals.PROCESSING_RESOURCE_SPECIFICATION__PROCESSING_RESOURCE_TYPE));
		featureMatchers.put(ResourceconfigurationPackage.Literals.NUMBER_OF_PARALLEL_PROCESSING_UNITS,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));

		// Resourcetype
		featureMatchers.put(ResourcetypePackage.Literals.PROCESSING_RESOURCE_TYPE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(ResourcetypePackage.Literals.COMMUNICATION_LINK_RESOURCE_TYPE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		featureMatchers.put(ResourcetypePackage.Literals.PASSIVE_RESOURCE_TYPE,
				features(CorePackage.Literals.NAMED_ELEMENT__NAME));
		// TODO: more to come
	}

	private static StructuralFeatureMatcher features(EStructuralFeature... f) {
		return new StructuralFeatureMatcher(f);
	}

	public static boolean matches(EObject o1, EObject o2) {
		if (o1 == null || o2 == null) {
			return false;
		}
		if (o1.getClass().equals(o2.getClass())) {
			StructuralFeatureMatcher matcher = featureMatchers.get(o1.eClass());
			if (matcher != null) {
				return matcher.matches(o1, o2);
			} else {
				defaultMatcher.matches(o1, o2);
			}
		}
		return false;
	}

}
