package tools.descartes.prisma.core.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.cdo.common.id.CDOIDUtil;
import org.eclipse.emf.cdo.eresource.CDOResourceFolder;
import org.eclipse.emf.cdo.eresource.CDOResourceNode;
import org.eclipse.emf.cdo.transaction.CDOTransaction;
import org.eclipse.emf.cdo.view.CDOAdapterPolicy;
import org.eclipse.emf.cdo.view.CDOQuery;
import org.eclipse.emf.cdo.view.CDOView;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.prisma.core.MessageBus;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.RepositoryConnection;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.Transaction.LockType;
import tools.descartes.prisma.core.scopes.Scope;
import tools.descartes.prisma.core.scopes.ScopeFactory;
import tools.descartes.prisma.core.scopes.SystemGlobalScope;

public class CDOModelRepository implements ModelRepository {

	private final RepositoryConnection connection;
	private CDOView readOnlyView;
	private final Map<String, Scope> scopes = new HashMap<>();
	private SystemGlobalScope systemScope;
	private final MessageBus messageBus;

	public CDOModelRepository(String host, String userName, String password, MessageBus messageBus) {
		this.connection = new RepositoryConnection(host);
		this.messageBus = messageBus;
	}

	@Override
	public void connect() {
		connection.open();
		this.readOnlyView = connection.startView();
		/*
		 * Important: To enable the notification of changes from other
		 * transactions/views change the subscription policy.
		 */
		this.readOnlyView.options().addChangeSubscriptionPolicy(CDOAdapterPolicy.CDO);
		this.systemScope = new SystemGlobalScope(this, messageBus);
	}

	public void disconnect() {
		try {
			connection.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public SystemGlobalScope getSystemScope() {
		return systemScope;
	}

	public Scope getScope(String path) {
		Scope scope = scopes.get(path);
		if (scope == null) {
			scope = createScope(path);
			if (scope != null) {
				scopes.put(path, scope);
			}
		}
		return scope;
	}

	private Scope createScope(String path) {
		return ScopeFactory.createScope(this, messageBus, systemScope, path);
	}

	@Override
	public void createFolders(Transaction transaction, String path) {
		CDOTransaction cdoTransaction = checkTransaction(transaction);

		String[] elements = path.split("/");
		if (elements.length < 1) {
			return;
		}
		CDOResourceNode[] nodes = cdoTransaction.getElements();
		CDOResourceFolder cur = null;
		for (CDOResourceNode curNode : nodes) {
			if (curNode.getName().equals(elements[0])) {
				cur = (CDOResourceFolder) curNode;
				break;
			}
		}
		if (cur == null) {
			transaction.lock(cdoTransaction.getRootResource(), LockType.WRITE);
			cur = cdoTransaction.createResourceFolder(elements[0]);
		}
		for (int i = 1; i < elements.length; i++) {
			CDOResourceFolder next = (CDOResourceFolder) cur.getNode(elements[i]);
			if (next == null) {
				transaction.lock(cur, LockType.WRITE);
				next = cur.addResourceFolder(elements[i]);
			}
			cur = next;
		}
	}

	@Override
	public Resource getResource(String path) {
		return readOnlyView.getResource(path);
	}

	@Override
	public Resource getOrCreateResource(Transaction transaction, String path) {
		CDOTransaction cdoTransaction = checkTransaction(transaction);
		int idx = path.lastIndexOf("/");
		String basePath = "";
		String name = path;
		if (idx >= 0) {
			basePath = path.substring(0, idx);
			name = path.substring(idx + 1);
		}
		CDOResourceFolder folder = cdoTransaction.getResourceFolder(basePath);
		CDOResourceNode node = folder.getNode(name);
		if (node != null) {
			return (Resource) node;
		} else {
			transaction.lock(folder, LockType.WRITE);
			return folder.addResource(name);
		}
	}

	@Override
	public List<String> listResources(String path) {
		CDOResourceFolder folder = readOnlyView.getResourceFolder(path);
		EList<CDOResourceNode> nodes = folder.getNodes();
		List<String> nodeNames = new ArrayList<String>(nodes.size());
		for (CDOResourceNode curNode : nodes) {
			nodeNames.add(curNode.getName());
		}
		return nodeNames;
	}

	@Override
	public boolean exists(String path) {
		CDOResourceFolder folder = null;
		if (!path.isEmpty()) {
			String[] elements = path.split("/");
			for (int i = 0; i < elements.length; i++) {
				List<CDOResourceNode> res = readOnlyView.queryResources(folder, elements[i], true);
				if (res.isEmpty()) {
					return false;
				}
				if (i == (elements.length - 1)) {
					return true;
				} else {
					CDOResourceNode node = res.get(0);
					if (node instanceof CDOResourceFolder) {
						folder = (CDOResourceFolder) node;
					} else {
						return false;
					}
				}
			}
		}
		return true;
	}

	@Override
	public String getInstanceID(Object object) {
		StringBuilder out = new StringBuilder();
		CDOIDUtil.write(out, ((CDOObject) object).cdoID());
		return out.toString();
	}

	@Override
	public Object getInstance(String id) {
		return readOnlyView.getObject(CDOIDUtil.read(id));
	}

	public Transaction createTransaction() {
		return new CDOTransactionWrapper(this, connection.startTransaction());
	}

	@Override
	public void waitForUpdate() {
		readOnlyView.waitForUpdate(readOnlyView.getLastUpdateTime() + 1);
	}

	@Override
	public <T> List<T> query(String query, Object context, Class<T> resultType) {
		CDOQuery cdoQuery = readOnlyView.createQuery("ocl", query, context);
		return cdoQuery.getResult(resultType);
	}

	protected CDOView getReadOnlyView() {
		return readOnlyView;
	}

	private CDOTransaction checkTransaction(Transaction transaction) {
		if (transaction == null) {
			throw new IllegalArgumentException("Transaction may not be null.");
		}
		return ((CDOTransactionWrapper) transaction).getWrappedTransaction();
	}

	/**
	 * We do not use the standard {@link EcoreUtil#copy(EObject)} function since
	 * we do not do a complete copy. Only attributes and non-containment
	 * references are copied to the new object.
	 * 
	 * @param obj
	 * @return
	 */
	private EObject copy(EObject obj) {
		EObject copy = EcoreUtil.create(obj.eClass());
		for (EStructuralFeature feature : obj.eClass().getEStructuralFeatures()) {
			if (feature.isChangeable() && !feature.isDerived()) {
				if (feature instanceof EAttribute) {
					if (obj.eIsSet(feature)) {
						if (feature.isMany()) {
							List<?> source = (List<?>) obj.eGet(feature);
							@SuppressWarnings("unchecked")
							List<Object> target = (List<Object>) copy.eGet(feature);
							if (source.isEmpty()) {
								target.clear();
							} else {
								target.addAll(source);
							}
						} else {
							copy.eSet(feature, obj.eGet(feature));
						}
					}
				}
			}
		}
		return copy;
	}
}
