package tools.descartes.prisma.core.merge;

import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.BranchAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.BranchProbabilities;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorPackage;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ObjectMatcher;
import tools.descartes.prisma.core.agent.AgentController;

public class ComponentInternalBehaviorMerger extends ObjectMerger {

	private static final String MERGE_IDENTIFIER = "[[PRISMA_MERGE]]";

	public ComponentInternalBehaviorMerger(AgentController agent) {
		super(agent);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * tools.descartes.prisma.core.merge.ObjectMerger#mergeTree(tools.descartes.
	 * prisma.core.Transaction, org.eclipse.emf.ecore.EObject,
	 * org.eclipse.emf.ecore.EObject, java.util.Set)
	 */
	@Override
	public void mergeTree(Transaction transaction, EObject base, EObject changed,
			Set<EStructuralFeature> changedFeatures) {
		// A special merging strategy is required for component internal
		// behaviors, since
		// they can contain an ordered list of actions. The ordering of these
		// actions need to
		// be preserved when merging two independently create component internal
		// behaviors
		ComponentInternalBehavior baseBehavior = (ComponentInternalBehavior) base;
		ComponentInternalBehavior changedBehavior = (ComponentInternalBehavior) changed;
		if (!baseBehavior.eIsSet(ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS)) {
			// Special case if in the current version the feature is not set
			EList<AbstractAction> changedList = changedBehavior.getActions();
			int size = changedList.size();
			for (int i = 0; i < size; i++) {
				AbstractAction mergeResult = (AbstractAction) mergeValue(
						ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS, null, changedList.get(i));
				baseBehavior.getActions().add(mergeResult);
				connectValue(transaction, ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS,
						changedList.get(i), mergeResult);
			}
		} else {
			int lastMatch = -1;
			Queue<AbstractAction> actionsToMerge = new LinkedList<>(changedBehavior.getActions());
			while (!actionsToMerge.isEmpty()) {
				AbstractAction curAction = actionsToMerge.peek();
				int match = findMatch(baseBehavior, lastMatch, curAction);
				if (match == (lastMatch + 1)) {
					// This is a direct match at the same position. It is enough
					// to just connect the two elements.
					connectValue(transaction, ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS,
							curAction, baseBehavior.getActions().get(match));
					actionsToMerge.poll();
					lastMatch = match;
				} else {
					// There is a conflict that needs to be resolved.
					List<AbstractAction> conflictingActions = new LinkedList<>();
					// Determine all actions that need to moved to a side branch
					// in order to resolve the conflict.
					int returnIdx = determineConflictingActions(baseBehavior, lastMatch, conflictingActions, actionsToMerge);
					boolean newMergeBranch = true;
					if (baseBehavior.getActions().isEmpty() || !isMergeBranch(baseBehavior.getActions().get(lastMatch + 1))) {
						newMergeBranch = true;
					} else {
						if (returnIdx == (lastMatch + 2)
								|| (returnIdx < 0 && baseBehavior.getActions().size() == (lastMatch + 2))) {
							newMergeBranch = false;
						}
					}

					if (newMergeBranch) {
						int end = (returnIdx < 0) ? baseBehavior.getActions().size() : returnIdx;
						insertAlternativeBranch(transaction, baseBehavior, lastMatch + 1, end, conflictingActions);
						lastMatch = returnIdx;
					} else {
						BranchAction mergeBranch = (BranchAction) baseBehavior.getActions().get(lastMatch + 1);
						ComponentInternalBehavior existingAlternative = null;
						for (ComponentInternalBehavior curBranch : mergeBranch.getBranches()) {
							if (!curBranch.getActions().isEmpty()) {
								if (ObjectMatcher.matches(curBranch.getActions().get(0), curAction)) {
									existingAlternative = curBranch;
									break;
								}
							}
						}
						if (existingAlternative == null) {
							existingAlternative = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
							for (AbstractAction curConflict : conflictingActions) {
								AbstractAction mergeResult = (AbstractAction) mergeValue(
										ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS, null,
										curConflict);
								existingAlternative.getActions().add(curConflict);
								connectValue(transaction,
										ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS,
										curConflict, mergeResult);
							}
							existingAlternative.getActions().addAll(conflictingActions);
							mergeBranch.getBranches().add(existingAlternative);
						} else {
							// TODO
						}
					}
				}
			}
		}
	}

	private void insertAlternativeBranch(Transaction transaction, ComponentInternalBehavior base, int conflictStart,
			int conflictEnd,
			List<AbstractAction> conflictingActions) {
		ComponentInternalBehavior newAlternative = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();

		// Add the conflicting actions from the model skeleton
		// to the new alternative branch
		for (AbstractAction curAction : conflictingActions) {
			AbstractAction mergeResult = (AbstractAction) mergeValue(
					ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS, null, curAction);
			newAlternative.getActions().add(curAction);
			connectValue(transaction, ServicebehaviorPackage.Literals.COMPONENT_INTERNAL_BEHAVIOR__ACTIONS, curAction,
					mergeResult);
		}

		// Now move the existing actions to another branch.
		ComponentInternalBehavior existingAlternative = ServicebehaviorFactory.eINSTANCE
				.createComponentInternalBehavior();
		for (int i = conflictStart; i < conflictEnd; i++) {
			existingAlternative.getActions().add(base.getActions().get(conflictStart));
		}
		BranchAction mergeAction = insertMergeBranch(base, conflictStart);
		mergeAction.getBranches().add(existingAlternative);
		mergeAction.getBranches().add(newAlternative);
	}

	private int determineConflictingActions(ComponentInternalBehavior base, int basePosition,
			List<AbstractAction> conflictingActions, Queue<AbstractAction> actionsToMerge) {
		while (!actionsToMerge.isEmpty()) {
			AbstractAction curAction = actionsToMerge.poll();
			int idx = findMatch(base, basePosition, curAction);
			if (idx >= 0) {
				return idx;
			} else {
				conflictingActions.add(curAction);
			}
		}
		return -1;
	}

	private int findMatch(ComponentInternalBehavior behavior, int startIdx, AbstractAction searchedAction) {
		int numActions = behavior.getActions().size();
		for (int i = (startIdx + 1); i < numActions; i++) {
			if (ObjectMatcher.matches(behavior.getActions().get(i), searchedAction)) {
				return i;
			}
		}
		return -1;
	}

	private boolean isMergeBranch(AbstractAction action) {
		if (action != null) {
			if (action instanceof BranchAction) {
				return ((BranchAction) action).getBranchProbabilities().getName().startsWith(MERGE_IDENTIFIER);
			}
		}
		return false;
	}

	private BranchAction insertMergeBranch(ComponentInternalBehavior behavior, int position) {
		AbstractAction action = null;
		if (behavior.getActions().size() > position) {
			action = behavior.getActions().get(position);
		}
		BranchAction mergeBranch;
		if (isMergeBranch(action)) {
			mergeBranch = (BranchAction) action;
		} else {
			mergeBranch = ServicebehaviorFactory.eINSTANCE.createBranchAction();
			BranchProbabilities probabilities = ServicebehaviorFactory.eINSTANCE.createBranchProbabilities();
			probabilities.setName(MERGE_IDENTIFIER);
			mergeBranch.setBranchProbabilities(probabilities);
			behavior.getActions().add(position, mergeBranch);
		}
		return mergeBranch;
	}

}
