package tools.descartes.prisma.core.owner;

import java.util.List;

import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.cdo.CDOState;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.agent.RemoveFilter;
import tools.descartes.prisma.model.scope.Agent;
import tools.descartes.prisma.model.scope.ObjectOwnership;
import tools.descartes.prisma.model.scope.ScopeConfiguration;

public class OwnershipManager {

	private final ModelRepository repository;
	private final ScopeConfiguration scopeConfiguration;
	private final Agent agent;

	public OwnershipManager(ModelRepository repository, ScopeConfiguration scopeConfiguration, Agent agent) {
		this.repository = repository;
		this.scopeConfiguration = scopeConfiguration;
		this.agent = agent;
	}

	public void takeOwnership(Transaction transaction, CDOObject object) {
		ObjectOwnership ownership = getOwnership(object);
		Agent agentInTransaction = transaction.attach(this.agent);
		if (ownership == null) {
			ScopeConfiguration dirtyScopeConfiguration = transaction.attach(scopeConfiguration);
			ownership = tools.descartes.prisma.model.scope.ScopeFactory.eINSTANCE.createObjectOwnership();
			ownership.setObject(object);
			ownership.getOwners().add(agentInTransaction);
			dirtyScopeConfiguration.getOwnedObjects().add(ownership);
		} else {
			ObjectOwnership ownershipInTransaction = transaction.attach(ownership);
			if (!ownershipInTransaction.getOwners().contains(agentInTransaction)) {
				ownershipInTransaction.getOwners().add(agentInTransaction);
			}
		}
	}

	public void releaseAllOwnerships(Transaction transaction, RemoveFilter filter) {
		Agent agentInTransaction = transaction.attach(agent);

		// Iterate through all objects and determine whether it should be
		// removed or not
		for (ObjectOwnership curOwnership : agentInTransaction.getOwnedObjects()) {
			if (curOwnership.getOwners().size() == 1) {
				// we are the only agent using this object, we may remove it
				// from the repository
				if (filter != null) {
					if (filter.isRemovedOnShutdown(curOwnership.getObject())) {
						// Remove this element completely
						EcoreUtil.remove(curOwnership.getObject());
						EcoreUtil.remove(curOwnership);
					}
				}
			}
		}

		// The remaining ownership are just deregistered, but the objects remain
		// in the repository
		agentInTransaction.getOwnedObjects().clear();
	}

	public boolean isShared(CDOObject object) {
		ObjectOwnership ownership = getOwnership(object);
		if (ownership == null) {
			// object not known --> it is not shared
			return false;
		} else {
			// there is more than 1 owner
			return ownership.getOwners().size() > 1;
		}
	}

	private ObjectOwnership getOwnership(CDOObject object) {
		if (object.cdoState() == CDOState.NEW) {
			// This object is not yet in the repository, so no one can own it.
			return null;
		} else {
			// IMPORTANT: use fully qualified type names in the query.
			// By default OCL uses the package of the context object (self) for
			// resolution which can be any package here.
			List<ObjectOwnership> res = repository.query(
					"scope::ObjectOwnership.allInstances()->select(o:scope::ObjectOwnership|o.object=self)",
					object.cdoID(),
					ObjectOwnership.class);
			if (res.isEmpty()) {
				return null;
			} else {
				return res.get(0);
			}
		}
	}

}
