package tools.descartes.prisma.parameterization;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.annotation.Resource;

import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCall;
import tools.descartes.librede.algorithm.EstimationAlgorithmFactory;
import tools.descartes.librede.algorithm.IConstrainedNonLinearOptimizationAlgorithm;
import tools.descartes.librede.algorithm.IEstimationAlgorithm;
import tools.descartes.librede.approach.AbstractEstimationApproach;
import tools.descartes.librede.configuration.ModelEntity;
import tools.descartes.librede.configuration.ResourceDemand;
import tools.descartes.librede.configuration.SchedulingStrategy;
import tools.descartes.librede.configuration.Service;
import tools.descartes.librede.configuration.Task;
import tools.descartes.librede.configuration.WorkloadDescription;
import tools.descartes.librede.linalg.LinAlg;
import tools.descartes.librede.linalg.Scalar;
import tools.descartes.librede.linalg.Vector;
import tools.descartes.librede.linalg.VectorBuilder;
import tools.descartes.librede.metrics.StandardMetrics;
import tools.descartes.librede.models.observation.IObservationModel;
import tools.descartes.librede.models.observation.IOutputWeightingFunction;
import tools.descartes.librede.models.observation.VectorObservationModel;
import tools.descartes.librede.models.observation.functions.AbstractDirectOutputFunction;
import tools.descartes.librede.models.observation.functions.IOutputFunction;
import tools.descartes.librede.models.observation.functions.ResponseTimeEquation;
import tools.descartes.librede.models.observation.functions.UtilizationLaw;
import tools.descartes.librede.models.state.ConstantStateModel;
import tools.descartes.librede.models.state.IStateModel;
import tools.descartes.librede.models.state.InvocationGraph;
import tools.descartes.librede.models.state.ConstantStateModel.Builder;
import tools.descartes.librede.models.state.constraints.IStateConstraint;
import tools.descartes.librede.models.state.constraints.NoRequestsBoundsConstraint;
import tools.descartes.librede.models.state.initial.WeightedTargetUtilizationInitializer;
import tools.descartes.librede.registry.Component;
import tools.descartes.librede.repository.IRepositoryCursor;
import tools.descartes.librede.repository.Query;
import tools.descartes.librede.repository.QueryBuilder;
import tools.descartes.librede.units.RequestRate;
import tools.descartes.librede.units.Time;

@Component(displayName = "Opt")
public class OptEstimationApproach extends AbstractEstimationApproach {	
	
	protected List<IStateModel<?>> deriveStateModelsIndividually(WorkloadDescription workload, IRepositoryCursor cursor) {
		for (Service s :  workload.getServices()) {
			List<Task> toBeRemoved = new ArrayList<>();
			for (Task t : s.getTasks()) {				
				if (t instanceof tools.descartes.librede.configuration.ExternalCall) {
					((tools.descartes.librede.configuration.ExternalCall) t).setCalledService(null);
					toBeRemoved.add(t);
				}
				if (t instanceof ResourceDemand) {
					if (((ResourceDemand)t).getResource().getSchedulingStrategy() == SchedulingStrategy.IS) {
						((ResourceDemand) t).setResource(null);
						toBeRemoved.add(t);
					}
				}
			}
			s.getTasks().removeAll(toBeRemoved);
		}
		
		ArrayList<IStateModel<?>> models = new ArrayList<IStateModel<?>>();
		for (tools.descartes.librede.configuration.Resource res : workload.getResources()) {
			if (res.getSchedulingStrategy() != SchedulingStrategy.IS) {
				Builder<IStateConstraint> builder = ConstantStateModel.constrainedModelBuilder();
				Set<Service> services = new HashSet<>();
				for (ResourceDemand demand : res.getDemands()) {
					builder.addConstraint(new NoRequestsBoundsConstraint(demand, cursor, 0, Double.POSITIVE_INFINITY));
					builder.addVariable(demand);
					services.add(demand.getService());
				}
				builder.setStateInitializer(new WeightedTargetUtilizationInitializer(0.5, cursor));
				builder.setInvocationGraph(new InvocationGraph(new ArrayList<>(services), cursor, getEstimationWindow()));
				models.add(builder.build());
			}
		}
		return models;
	}

	protected List<IStateModel<?>> deriveStateModels(WorkloadDescription workload, IRepositoryCursor cursor) {
//		return deriveStateModelsIndividually(workload, cursor);
		Builder<IStateConstraint> builder = ConstantStateModel.constrainedModelBuilder();
		Set<Service> services = new HashSet<>();
		for (tools.descartes.librede.configuration.Resource res : workload.getResources()) {
			for (ResourceDemand demand : res.getDemands()) {
				builder.addConstraint(new NoRequestsBoundsConstraint(demand, cursor, 0, Double.POSITIVE_INFINITY));
				builder.addVariable(demand);
				services.add(demand.getService());
			}
		}		
		builder.setStateInitializer(new WeightedTargetUtilizationInitializer(0.5, cursor));
		builder.setInvocationGraph(new InvocationGraph(workload.getServices(), cursor, getEstimationWindow()));
		return Collections.singletonList(builder.build());
	}
	
	protected IObservationModel<?, ?> deriveObservationModel(IStateModel<?> stateModel, IRepositoryCursor cursor) {
		ThroughputWeights weightsFunction = new ThroughputWeights();
		VectorObservationModel<IOutputFunction> observationModel = new VectorObservationModel<IOutputFunction>(weightsFunction);
//		VectorObservationModel<IOutputFunction> observationModel = new VectorObservationModel<IOutputFunction>();
		List<ModelEntity> entities = new LinkedList<>();
		for (int i = 0; i < getEstimationWindow(); i++) {
			for (Service service : stateModel.getUserServices()) {
				// Current assumption is that services which are not called by others
				// are the system entry services. Since we look at the end to end
				// response time, we only include these services to the estimation.
				// The remaining services are referenced from these system entry services.
				if (service.getIncomingCalls().isEmpty()) {
					ResponseTimeEquation func = new ResponseTimeEquation(stateModel, cursor, service, true, i);
					observationModel.addOutputFunction(func);
					entities.add(service);
				}
			}
			for (tools.descartes.librede.configuration.Resource resource : stateModel.getResources()) {
				if (resource.getSchedulingStrategy() != SchedulingStrategy.IS) {
					UtilizationLaw func = new UtilizationLaw(stateModel, cursor, resource, i);
					observationModel.addOutputFunction(func);
					entities.add(resource);
				} else {
					if (i == 0) {
						for (Service curService : resource.getAccessingServices()) {
							observationModel.addOutputFunction(new DelayMinimizationLaw(stateModel, cursor, resource, curService));
							entities.add(resource);
						}
					}
				}
			}
		}
		weightsFunction.setEntities(cursor, entities);			
		return observationModel;
	}
	
	protected IEstimationAlgorithm getEstimationAlgorithm(EstimationAlgorithmFactory factory) {
		return factory.createInstance(IConstrainedNonLinearOptimizationAlgorithm.class);
	}
	
	private static class DelayMinimizationLaw extends AbstractDirectOutputFunction {
		
		private Query<Scalar, Time> rt;

		public DelayMinimizationLaw(IStateModel<? extends IStateConstraint> stateModel, IRepositoryCursor cursor, tools.descartes.librede.configuration.Resource res_i, Service cls_r) {
			super(stateModel, res_i, cls_r, 0);
			rt = QueryBuilder.select(StandardMetrics.RESPONSE_TIME).in(Time.SECONDS).forService(cls_r).average().using(cursor);
		}

		@Override
		public boolean isApplicable(List<String> messages) {
			return true;
		}

		@Override
		public boolean hasData() {
			return true;
		}

		@Override
		public double getObservedOutput() {
			return 0.0;
		}

		@Override
		public double getFactor() {
			double curRt = rt.execute().getValue();
			if (curRt == 0 || Double.isNaN(curRt)) {
				return 0.0;
			}
			return 1 / rt.execute().getValue();
//			return 0.0;
		}		
	}
	
	private class ThroughputWeights implements IOutputWeightingFunction {

		private Query<Vector, RequestRate> tputQuery;
		private List<ModelEntity> entities;
		
		public void setEntities(IRepositoryCursor cursor, List<ModelEntity> entities) {
			this.entities = entities;
			Set<Service> services = new HashSet<>();
			for (ModelEntity curEntity : entities) {
				if (curEntity instanceof Service) {
					services.add((Service)curEntity);
				}
			}
			tputQuery = QueryBuilder.select(StandardMetrics.THROUGHPUT).in(RequestRate.REQ_PER_SECOND).forServices(services).average().using(cursor);
		}

		@Override
		public Vector getOutputWheights() {
			Vector tput = tputQuery.execute();
			double sumTput = LinAlg.sum(tput).get(0);
			VectorBuilder weights = VectorBuilder.create(entities.size());
			for (ModelEntity curEntity : entities) {
				if (curEntity instanceof Service) {
					if (sumTput > 0.0) {
						weights.add(tput.get(tputQuery.indexOf(curEntity)) / sumTput);
					} else {
						weights.add(0);
					}
				} else if (curEntity instanceof tools.descartes.librede.configuration.Resource) {
					tools.descartes.librede.configuration.Resource curRes = (tools.descartes.librede.configuration.Resource)curEntity;
					if (curRes.getSchedulingStrategy() == SchedulingStrategy.IS) {
						weights.add(getEstimationWindow());
					} else {
						weights.add(1);
					}
				}
			}
			return weights.toVector();
		}
		
	}

}
